/**********
 * Copyright (c) 2003-2004 Greg Parker.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY GREG PARKER ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **********/

#include "includes.h"
#include "vt100.h"
#include "formutils.h"
#include "data/prefs.h"
#include "rsrc/rsrc.h"
#include "ssh/ssh.h"

#define PUTTY_DO_GLOBALS
#include "putty.h"


// Default color set
// BOLD is more saturated, except for white and black themselves
static RGBColorType colors[] = {
    {0, 0x00,0x00,0x00}, {0, 0x22,0x22,0x22}, 
    {0, 0xdd,0x00,0x00}, {0, 0xff,0x00,0x00}, 
    {0, 0x00,0xdd,0x00}, {0, 0x00,0xff,0x00}, 
    {0, 0xdd,0xdd,0x00}, {0, 0xff,0xff,0x00}, 
    {0, 0x00,0x00,0xdd}, {0, 0x00,0x00,0xff}, 
    {0, 0xdd,0x00,0xdd}, {0, 0xff,0x00,0xff}, 
    {0, 0x00,0xdd,0xdd}, {0, 0x00,0xff,0xff}, 
    {0, 0xff,0xff,0xff}, {0, 0xdd,0xdd,0xdd}, 
};

static unsigned int DefaultForeColor = defaultForeColor;
static unsigned int DefaultBackColor = defaultBackColor;


void WIN_read_default_colors(void)
{
    DefaultForeColor = PrefsGetInt(prefTerminalForeColor, defaultForeColor);
    if (DefaultForeColor >= 8) DefaultForeColor = defaultForeColor;
    DefaultBackColor = PrefsGetInt(prefTerminalBackColor, defaultBackColor);
    if (DefaultBackColor >= 8) DefaultBackColor = defaultBackColor;
}

extern display *gdd;

void set_palm_color(int fg_color, int bg_color)
{
    WinSetForeColorRGB(&colors[fg_color], NULL);
    WinSetTextColorRGB(&colors[fg_color], NULL);
    WinSetBackColorRGB(&colors[bg_color], NULL);
}

FontID font_for_screen(int fontPref)
{
    UInt32 density;
    Boolean lowres;

    WinScreenGetAttribute(winScreenDensity, &density);

    switch (density) {
    case kDensityDouble:
    case kDensityTriple:
    case kDensityQuadruple:
        lowres = false;
        break;
    case kDensityLow:
    case kDensityOneAndAHalf:
    default:
        lowres = true;
        break;
    }

    switch (fontPref) {
    case font6x10: return lowres ? MediumFontSingleID : MediumFontDoubleID;
    case font4x6: 
    default: return lowres ? NanoFontSingleID : NanoFontDoubleID; 
    }
}

void WIN_request_resize(int w, int h)
{
    // server-side size change - not supported
    // Disabled with PuTTY's cfg.no_remote_resize
}


static int SectRect(RectangleType a, RectangleType b)
{
    RectangleType c;
    RctGetIntersection(&a, &b, &c);
    return (c.extent.x == 0  ||  c.extent.y == 0);
}

/*
 * Draw a line of text in the window, at given character
 * coordinates, in given attributes.
 *
 * We are allowed to fiddle with the contents of `text'.
 * 
 * All calculations below are NATIVE-RESOLUTION
 */
void WIN_do_text(Context ctx, int x, int y, char *text, int len,
                 unsigned long attr, int lattr)
{
#define LEFT(rr) ((rr).topLeft.x)
#define TOP(rr) ((rr).topLeft.y)
#define RIGHT(rr) ((rr).topLeft.x + (rr).extent.x)
#define BOTTOM(rr) ((rr).topLeft.y + (rr).extent.y)

    display *dd = (display *)ctx;

    // fixme implement
    int fg_color;
    int bg_color;
    int font_width = dd->font_width;
    int font_height = dd->font_height;

    RectangleType r, big, fill;
    RectangleType closeBox;
    int closeBoxInvalid = 0;

    closeBox.topLeft.x = RIGHT(dd->gadgetBounds) - dd->closeBoxSize;
    closeBox.topLeft.y = TOP(dd->gadgetBounds);
    closeBox.extent.x = dd->closeBoxSize;
    closeBox.extent.y = dd->closeBoxSize;

    r.topLeft.x = dd->textBounds.topLeft.x + x * font_width;
    r.topLeft.y = dd->textBounds.topLeft.y + y * font_height;
    r.extent.x = len * font_width;
    r.extent.y = font_height;

    fg_color = (attr & ATTR_FGMASK) >> ATTR_FGSHIFT;
    bg_color = (attr & ATTR_BGMASK) >> ATTR_BGSHIFT;

    // handle default colors
    if (fg_color == (ATTR_DEFFG >> ATTR_FGSHIFT)) fg_color = DefaultForeColor;
    if (bg_color == (ATTR_DEFBG >> ATTR_BGSHIFT)) bg_color = DefaultBackColor;

    // translate into indexes into colors[]
    fg_color *= 2;
    bg_color *= 2;

    if (attr & ATTR_REVERSE) {
        int temp = fg_color;
        fg_color = bg_color;
        bg_color = temp;
    }

    // fixme can do real bold and underline with bigger fonts
    if ((attr & ATTR_BOLD)  ||  (attr & ATTR_UNDER)) {
        fg_color++;
    }
    // fixme allow optional real blink
    if (attr & ATTR_BLINK) {
        bg_color++;
    }

    set_palm_color(fg_color, bg_color);
    WinDrawChars(text, len, r.topLeft.x, r.topLeft.y);
    if (SectRect(r, closeBox)) closeBoxInvalid = 1;

    // If the drawn rect extends to any edge of the text area, fill the 
    // gap between the text area and the gadget bounds with the 
    // background color.
    big = r;
    if (LEFT(r) == LEFT(dd->textBounds)) {
        big.topLeft.x = LEFT(dd->gadgetBounds);
        big.extent.x += LEFT(dd->textBounds) - LEFT(dd->gadgetBounds);
    }
    if (TOP(r) == TOP(dd->textBounds)) {
        big.topLeft.y = TOP(dd->gadgetBounds);
        big.extent.y += TOP(dd->textBounds) - TOP(dd->gadgetBounds);
    }
    if (RIGHT(r) == RIGHT(dd->textBounds)) {
        big.extent.x += RIGHT(dd->gadgetBounds) - RIGHT(dd->textBounds);
    }
    if (BOTTOM(r) == BOTTOM(dd->textBounds)) {
        big.extent.y += BOTTOM(dd->gadgetBounds) - BOTTOM(dd->textBounds);
    }

    // Fill any gaps outside big and inside r
    if (LEFT(big) != LEFT(r)) {
        fill.topLeft.x = LEFT(big);
        fill.topLeft.y = TOP(big);
        fill.extent.x = LEFT(r) - LEFT(big);
        fill.extent.y = big.extent.y;
        WinEraseRectangle(&fill, 0);
        if (SectRect(fill, closeBox)) closeBoxInvalid = 1;
    }
    if (TOP(big) != TOP(r)) {
        fill.topLeft.x = LEFT(big);
        fill.topLeft.y = TOP(big);
        fill.extent.x = big.extent.x;
        fill.extent.y = TOP(r) - TOP(big);
        WinEraseRectangle(&fill, 0);
        if (SectRect(fill, closeBox)) closeBoxInvalid = 1;
    }
    if (RIGHT(big) != RIGHT(r)) {
        fill.topLeft.x = RIGHT(r);
        fill.topLeft.y = TOP(big);
        fill.extent.x = RIGHT(big) - RIGHT(r);
        fill.extent.y = big.extent.y;
        WinEraseRectangle(&fill, 0);
        if (SectRect(fill, closeBox)) closeBoxInvalid = 1;
    }
    if (BOTTOM(big) != BOTTOM(r)) {
        fill.topLeft.x = LEFT(big);
        fill.topLeft.y = BOTTOM(r);
        fill.extent.x = big.extent.x;
        fill.extent.y = BOTTOM(big) - BOTTOM(r);
        WinEraseRectangle(&fill, 0);
        if (SectRect(fill, closeBox)) closeBoxInvalid = 1;
    }

    // cursor
    // fixme better cursor, box cursor, etc
    if (attr & (TATTR_ACTCURS | TATTR_PASCURS | TATTR_RIGHTCURS)) {
        WinInvertRectangle(&r, 0);
    }

    if (closeBoxInvalid) {
        EventType e = {0};
        e.eType = usrDrawCloseBoxEvent;
        EvtAddUniqueEventToQueue(&e, 0, true);
    }
}

void WIN_do_cursor(Context ctx, int x, int y, char *text, int len,
                   unsigned long attr, int lattr)
{
    // fixme implement
    // fixme is active/passive/right necessary?
    WIN_do_text(ctx, x, y, text, len, attr, lattr);
}

int WIN_CharWidth(Context ctx, int uc) 
{
    return 1;
    // fixme implement for fonts with double-wide chars
}

void WIN_set_title(char *title)
{
    // fixme change user@host to this title?
}

void WIN_set_icon(char *title)
{
    // fixme probably nowhere to put this
}

void WIN_set_sbar(int total, int start, int page)
{
    Boolean newEnabled;
    Boolean oldEnabled;

    // fixme don't crash if terminal form is not in front
    if (FrmGetActiveFormID() != TerminalFormID) return; 

    // enable or disable up arrows
    newEnabled = (start != 0);
    oldEnabled = PrvGetControlEnabled(TerminalRScrollLineUpButtonID);
    if (newEnabled != oldEnabled) {
        PrvSetControlEnabled(TerminalRScrollLineUpButtonID, newEnabled);
        PrvSetControlEnabled(TerminalRScrollPageUpButtonID, newEnabled);
        if (newEnabled) {
            PrvSetControlGraphics(TerminalRScrollLineUpButtonID, TerminalRScrollLineUpBitmapID, NULL);
            PrvSetControlGraphics(TerminalRScrollPageUpButtonID, TerminalRScrollPageUpBitmapID, NULL);
        } else {
            PrvSetControlGraphics(TerminalRScrollLineUpButtonID, TerminalRScrollLineUpDisabledBitmapID, NULL);
            PrvSetControlGraphics(TerminalRScrollPageUpButtonID, TerminalRScrollPageUpDisabledBitmapID, NULL);
        }
    }

    // enable or disable down arrows
    newEnabled = (total > start + page);
    oldEnabled = PrvGetControlEnabled(TerminalRScrollLineDownButtonID);
    if (newEnabled != oldEnabled) {
        PrvSetControlEnabled(TerminalRScrollLineDownButtonID, newEnabled);
        PrvSetControlEnabled(TerminalRScrollPageDownButtonID, newEnabled);
        if (newEnabled) {
            PrvSetControlGraphics(TerminalRScrollLineDownButtonID, TerminalRScrollLineDownBitmapID, 0);
            PrvSetControlGraphics(TerminalRScrollPageDownButtonID, TerminalRScrollPageDownBitmapID, 0);
        } else {
            PrvSetControlGraphics(TerminalRScrollLineDownButtonID, TerminalRScrollLineDownDisabledBitmapID, 0);
            PrvSetControlGraphics(TerminalRScrollPageDownButtonID, TerminalRScrollPageDownDisabledBitmapID, 0);
        }
    }
}

Context WIN_get_ctx(void)
{
    // fixme return NULL if we aren't supposed to draw anything now
    if (gdd->inactive) return NULL;

    WinPushDrawState();
    WinSetCoordinateSystem(kCoordinatesNative);

    FntSetFont(gdd->fontID);

    return gdd;
}

void WIN_free_ctx(Context ctx)
{
    if (!ctx) return;
    WinPopDrawState();
}

void WIN_palette_set(int n, int r, int g, int b)
{
    // fixme Linux foo
}

void WIN_palette_reset(void)
{
    // fixme
}

void WIN_write_aclip(char *data, UInt16 len, int must_deselect)
{
    // fixme copy&paste implementation
}

/*
 * Note: unlike write_aclip() this will not append a nul.
 */
void WIN_write_clip(wchar_t * data, UInt16 len, int must_deselect)
{
    // wchar_t bad
    char *newdata;
    unsigned int i;
    // last byte of data is nul char - skip it
    // convert newlines - Palm uses LF
    len--;
    newdata = smalloc(len);
    for (i = 0; i < len; i++) {
        if (data[i] == 13) newdata[i] = 10;
        else newdata[i] = (char)data[i];
    }

    ClipboardAddItem(clipboardText, newdata, len);
    sfree(newdata);
}

void WIN_get_clip(wchar_t ** p, UInt16 *len)
{
    MemHandle clipH;
    char *clipP;
    static wchar_t *wclipP = NULL;

    if (p) {
        clipH = ClipboardGetItem(clipboardText, len);
        if (clipH) {
            UInt16 i;
            clipP = MemHandleLock(clipH);
            wclipP = smalloc(*len * sizeof(wchar_t));
            for (i = 0; i < *len; i++) {
                wclipP[i] = (wchar_t)clipP[i];
            }
            *p = wclipP;
            MemHandleUnlock(clipH);
        } else {
            // no clipboard to paste
            *p = NULL;
            wclipP = NULL;
        }
    } else {
        // clear previously retrieved clip
        if (wclipP) {
            sfree(wclipP);
            wclipP = NULL;
        }
    }
}

void WIN_set_raw_mouse_mode(int activate)
{
    // fixme does Palm raw mouse mode mean anything
    // disable with cfg.no_mouse_rep
}

Mouse_Button WIN_translate_button(Mouse_Button button)
{
    // transform left/middle/right to select/extend/paste
    // Only one "mouse button" on Palm, and it selects.

    return MBT_SELECT;
}

void WIN_beep(int mode)
{
    // ignore mode - use prefs instead
    // fixme what if there's a dialog on screen?
    
    uint32_t beep = PrefsGetInt(prefBellBeep, defaultBellBeep);
    uint32_t flash = PrefsGetInt(prefBellFlash, defaultBellFlash);
    uint32_t flashEnd;
    display *dd = WIN_get_ctx();
    static int i = 0;

    if (dd) {
        if (flash) {
            WinInvertRectangle(&dd->gadgetBounds, 0);
            // make sure flash stays visible for 1/10 sec
            flashEnd = TimGetTicks() + (SysTicksPerSecond() / 10);
        }
        if (beep) {
            // beep at 400 Hz for 100 ms
            Err err;
            SndCommandType snd;
            UInt16 volume = PrefGetPreference(prefGameSoundVolume);
            snd.cmd = sndCmdFrqOn; // asynchronous
            snd.param1 = 400; // frequency in Hz
            snd.param2 = 100; // duration in ms
            snd.param3 = volume; // volume [0..sndMaxAmp]
            
            err = SndDoCmd(NULL, &snd, false);
        }
        if (flash) {
            // fixme make this asynchronous - use PuTTY's native handling?
            uint32_t now = TimGetTicks();
            if (now < flashEnd) SysTaskDelay(flashEnd - now);
            WinInvertRectangle(&dd->gadgetBounds, 0);
        }
        
        WIN_free_ctx(dd);
    }
}

void WIN_sys_cursor(int x, int y)
{
    // no system cursor
    // Only used on Windows for the benefit of screen-readers
}

void WIN_set_iconic(int iconic)
{
    // server-side iconification request
    // No such thing as an iconic Palm window
}

void WIN_move_window(int x, int y)
{
    // server-side move request
    // No such thing as a moved Palm window
}

void WIN_set_zorder(int top)
{
    // server-side restack request
    // No such thing as Palm z-order
}

void WIN_refresh_window(void)
{
    // server-side redraw request
    // fixme
}

void WIN_set_zoomed(int zoomed)
{
    // server-side window zoom request
    // No such thing as a Palm zoomed window
}

int WIN_is_iconic(void)
{
    // No such thing as an iconic Palm window
    return 0;
}

void WIN_get_window_pos(int *x, int *y)
{
    // fixme in pixels?
    *x = 0;
    *y = 0;
}

void WIN_get_window_pixels(int *x, int *y)
{
    // fixme
    *x = 160;
    *y = 120;
}

char *WIN_get_window_title(int icon)
{
    // fixme
    return "fixme window title";
}


// PuTTY ticks are milliseconds
unsigned long WIN_GetTickCount(void)
{
    return TimGetTicks() / SysTicksPerSecond() * 1000L;
}


unsigned long WIN_GetCaretBlinkTime(void)
{
    return 500; // fixme
}


// fixme no printing
printer_enum *printer_start_enum(int *nprinters)
{
    *nprinters = 0;
    return 0;
}
char *printer_get_name(printer_enum *a, int b)
{
    return NULL;
}
void printer_finish_enum(printer_enum *c)
{
}
printer_job *printer_start_job(char *printer)
{
    return NULL;
}
void printer_job_data(printer_job *d, void *e, int f)
{
}
void printer_finish_job(printer_job *g)
{
}


void ldisc_send(char *buf, int len, int interactive)
{
    if (buf && len > 0) {
        ssh_bytes(ssh_get_current(), buf, len);
    }
}
